using System;
using System.Diagnostics;
using System.IO;
using System.Runtime.InteropServices;
using System.Text;
using System.Windows.Forms;
using mshtml;
using System.ComponentModel;
using System.Drawing;

namespace onlyconnect
{
	
	/// <summary>
	/// Implements MSHTML as active document in a control
	/// With thanks to Lutz Roeder
	/// 
	/// also thanks to Steven Wood for the HTML Event handling idea
	/// and code.
	/// 
	/// and thanks to Christopher Slee for the region marking
	/// 
	/// thanks to Jamie Hancock for many suggestions and improvements
	/// 
	/// thanks to Nikhil Kothari for some new COM definitions and 
	/// mshtml insights, see: http://www.nikhilk.net/Entry.aspx?id=11
	/// 
	/// Comments, improvements, suggestions - 
	/// email tim@itwriting.com
	/// or visit the messageboard at http://www.itwriting.com/htmleditor.php
	/// </summary>
	/// 
	/// Basic Reference:
	/// Activating the MSHTML Editor  Internet Development Index 
	/// --------------------------------------------------------------------------------
	/// This topic describes how to activate the MSHTML Editor from 
	/// Microsoft Visual C++, Microsoft Visual Basic, Microsoft JScript, and Visual Basic Scripting Edition (VBScript). 
	/// The Editor provides a rich set of capabilities and serves as a fine 
	/// "what you see is what you get" (WYSIWYG) HTML editing environment. 
	/// In addition, the Editor includes the ability to customize its behavior. 
	/// For more information, see Related Topics at the end of this topic.


	public class HtmlEditor : Control, IPropertyNotifySink
	{
		#region class variables and constructor

		private UCOMIConnectionPoint icp;
		private int cookie = -1;
		private string mReadyState = string.Empty;
		private bool bLoadDocumentWhenReady = false;
		private bool bSetComposeSettingsWhenReady = false;
		private int iLoadAttempts = 0;
		private EncodingType mDocumentEncoding = EncodingType.WindowsCurrent;
		private bool mIsWin98 = false; 
		private bool mAlwaysLoadAnsi = false;
		internal bool mLinksNewWindow = true;
		internal bool bNeedsActivation = false;

		HtmlSite site;
		String url = String.Empty;
		String sDocument = String.Empty;
		internal bool mDesignMode = false;
		private bool mIsContextMenuEnabled = true;
		private string 	mOptionKeyPath = null;
		internal bool mAllowActivation = true;
		internal  HTMLDocument m_htmldoc = null;
		internal IHTMLElement mcurrentElement = null;
		private IntPtr mDocumentHandle = IntPtr.Zero;
		private bool mCreating = true;

		private NativeWindow mNativeWindow = null;
		private HandleWndProc mNativeDocWindow = null;
		private System.Windows.Forms.Timer ActivateTimer;
		private System.ComponentModel.IContainer components;
		private ComposeSettings mComposeSettings = null; 

		internal bool mEnableActiveContent = true;

		public HtmlEditor() : base()
		{
			//initialize components
			InitializeComponent();
			
			//Detect Windows version
			mIsWin98 = (System.Environment.OSVersion.Platform == PlatformID.Win32Windows);
            				
			//force creation of handle, needed to host mshtml
			this.CreateControl();

			//see OnHandleCreated for purpose of mCreating
			mCreating = false;
		
		}

		#endregion class variables and constructor

		#region Invokes for external key presses etc. from the document.

		internal void InvokeOnMouseDown(MouseEventArgs e) 
		{
			this.OnMouseDown(e);
		}

		internal void InvokeOnDoubleClick() 
		{
			this.OnDoubleClick(EventArgs.Empty);
		}

		internal void InvokeWndProc(ref Message msg) 
		{
			Message message = Message.Create(this.Handle, msg.Msg, msg.WParam, msg.LParam);
			this.WndProc(ref message);
		}
		#endregion

		#region Overrides of container control methods ===============

		#region Override Handle events ===============

		protected override void OnHandleCreated(EventArgs e)
		{
			//earlier versions of HtmlEditor always init mshtml
			//here. However, according to BoundsChecker that 
			//causes a memory overrun. It also means that the 
			//document is passed a rect with zero bounds, as 
			//the control has not yet been sized.
			
			//Therefore, mshtml is now initialized for the first
			//time in OnParentChanged. In most cases, that will
			//be when the designer-generated code adds the control
			//to the form's controls collection.

			if (! mCreating)
			{
				initMshtml();
			}
		}

		protected override void OnHandleDestroyed(EventArgs e)
		{
			CleanupControl();
			base.OnHandleDestroyed(e);
		}


		#endregion Override Handle events

		public override Boolean PreProcessMessage(ref Message message)
		{
			return false; //handle elsewhere for the moment
			
		}

		override protected void WndProc(ref Message m)
		{
			switch (m.Msg) 
			{
				case ComSupport.WM_MOUSEACTIVATE:
				{
					if (site != null)
					{
	
						if (this.DesignMode) 
						{
							return;
						}
			
				
						IntPtr fromHandle = ComSupport.GetFocus();

						this.mDocumentHandle  = site.DocumentHandle;

						if ((! this.Focused) & (fromHandle != this.Handle) & (fromHandle != this.mDocumentHandle) )
						{
							ComSupport.SendMessage(this.Handle,ComSupport.WM_SETFOCUS,(int)fromHandle,0);
						}

					}
					break;
				}

				case ComSupport.WM_KEYUP:
					if (ComSupport.GetKeyState((int)Keys.ControlKey) < 0)  
					{
						Keys k = (Keys)m.WParam.ToInt32();
						doShortCut(k);
					}
					break;
			}

			base.WndProc(ref m);

		}


		protected override void OnParentChanged( System.EventArgs e)
		{
			if (site == null)
			{
				initMshtml();
			}

			base.OnParentChanged(e);
		}

		protected override void OnVisibleChanged(EventArgs e)
		{
			if (this.Visible)
			{
				if (this.bNeedsActivation)
				{
					this.ActivateTimer.Enabled = true;
				}
			}
			base.OnVisibleChanged (e);
		}


		protected override void OnPaint(PaintEventArgs pevent)
		{
			if (this.DesignMode)
			{
				pevent.Graphics.FillRectangle(System.Drawing.Brushes.White ,this.ClientRectangle);
			}
			
			base.OnPaint(pevent);
		}


		#endregion Overrides of container control methods ===============

		#region Override WndProc of mshtml 

		internal void setupWndProc()
		{

			if (this.mNativeDocWindow != null)
			{
				this.mNativeDocWindow.ReleaseHandle();
			}

			this.mNativeDocWindow = new HandleWndProc();
			this.mNativeDocWindow.thecontrol = this;
			this.mNativeDocWindow.AssignHandle(this.site.DocumentHandle);
		}

		internal void releaseWndProc()
		{
			if (this.mNativeDocWindow != null)
			{
				this.mNativeDocWindow.ReleaseHandle();
				this.mNativeDocWindow = null;
			}
		}

		#endregion Override WndProc of mshtml

		#region initialization routines

		void initMshtml()
		{

			//don't create in design mode
			if (this.DesignMode)
			{
				return;
			}

			if (this.isCreated)
			{
				return; //can't create if already created
			}

			// force creating Host handle since we need it to parent MSHTML
			if (! this.IsHandleCreated)
			{
				IntPtr hostHandle = Handle;
			}

			//get the NativeWindow so we hold onto the handle
			//this.mNativeWindow = NativeWindow.FromHandle(Handle);

			site = new HtmlSite(this);
			Debug.WriteLine("Site created");
			site.CreateDocument();
			Debug.WriteLine("Document created");
			
			if (this.Visible) 
			{
				site.ActivateDocument();
			}
			else
			{
			 this.bNeedsActivation = true;
				this.ActivateTimer.Enabled = true;
			}

			this.m_htmldoc = (HTMLDocument) site.Document;
	
			if (this.mDesignMode)
			{
				this.m_htmldoc.designMode = "On";

		}

			SetDocEvents();

			if (url != String.Empty) 
			{
				LoadUrl(url);
			}
			else
			{
				//always initialize with at least the blank url
				LoadUrl("About:Blank");
			}

			if (sDocument != String.Empty) 
			{
				LoadDocument(sDocument);
			}
			else
			{
				if (this.mDesignMode)
				{
					LoadDocument("<html></html>");
				}
			}

			if (this.mDesignMode & this.DefaultComposeSettings.Enabled)
			{
				this.setDefaultFont();
			}
		
		}

		
		private void InitializeComponent()
		{
			this.components = new System.ComponentModel.Container();
			this.ActivateTimer = new System.Windows.Forms.Timer(this.components);
			// 
			// ActivateTimer
			// 
			this.ActivateTimer.Interval = 1;
			this.ActivateTimer.Tick += new System.EventHandler(this.ActivateTimer_Tick);

		}


		#endregion initialization routines

		#region Dispose and cleanup routines

		/// <summary>
		/// Clean up any resources being used.
		/// </summary>
		protected override void Dispose( bool disposing )
		{
			
		
			
			if( disposing )
			{

				if (components != null) 
				{
					components.Dispose();
				}

				if (cookie != -1)
				{
					icp.Unadvise(cookie);
					cookie = -1;
				}

				IntPtr ptr = Marshal.GetIDispatchForObject(this);
				int i = Marshal.Release(ptr);
				while (i > 0)
				{
					i = Marshal.Release(ptr);
				}

				if (this.site != null)
				{
					this.site.Dispose();
				}

				if (this.mNativeWindow != null)
				{
					this.mNativeWindow.ReleaseHandle();
				}

			}
			base.Dispose( disposing );
		}

		public void CleanupControl()
		{
			if (isCreated)
			{
				if (cookie != -1)
				{
					icp.Unadvise(cookie);
					cookie = -1;
				}

				site.CloseDocument();
				site.Dispose();
				site = null;
				this.m_htmldoc = null;

			}
		}

		#endregion Dispose and cleanup routines

		#region Declare Events ====================================================

		//declare the events
		public event HtmlNavigateEventHandler Navigate;
		public event BeforeNavigateEventHandler BeforeNavigate;
		public event ReadyStateChangedHandler ReadyStateChanged;
		public event UpdateUIHandler UpdateUI;
		public event HtmlKeyPressHandler HtmlKeyPress; 
		public event HtmlEventHandler HtmlEvent;
		public event BeforeShortcutEventHandler BeforeShortcut;
		#endregion Declare Events ===========

		#region Invoke Events =========================================================
		//invoke the UpdateUI event
		public void InvokeUpdateUI(mshtml.IHTMLElement ae)
		{
			if (UpdateUI != null)
			{
				HtmlUpdateUIEventArgs ea = new HtmlUpdateUIEventArgs();
				ea.currentElement = ae;
				UpdateUI(this, ea);
			}
		}

		//invoke the ReadyStateChanged event
		public void InvokeReadyStateChanged(String newReadyState)
		{
			if (ReadyStateChanged != null)
			{
				ReadyStateChangedEventArgs ea = new ReadyStateChangedEventArgs(newReadyState);
				ReadyStateChanged(this, ea);
			}
		}
          
		public void InvokeHtmlKeyPress(ref mshtml.IHTMLEventObj eobj)
		{
			if (HtmlKeyPress != null)
			{
				HtmlKeyPressEventArgs ea = new HtmlKeyPressEventArgs(ref eobj);
				HtmlKeyPress(this, ea);
			}
		}

		[DispId(0)]
		public void InvokeHtmlEvent()
		{
			if (HtmlEvent != null)
			{
				// Get the event.
				mshtml.IHTMLEventObj pobjEvent = ((mshtml.IHTMLDocument2)
					m_htmldoc).parentWindow.@event;
				HtmlEventArgs ea = new HtmlEventArgs(pobjEvent);
				HtmlEvent(this, ea);
			}
			return;
		}
		
		public void InvokeNavigate(String target)
		{
			
			if (Navigate != null) Navigate(this, new HtmlNavigateEventArgs(target));
		}

		/// <summary>
		/// Fires the BeforeShortcut event. Handle this event to pre-process or cancel the
		/// HtmlEditor's shortcut events.
		/// </summary>
		/// <param name="e">Cancelable event args</param>
		public void OnBeforeShortcut(onlyconnect.BeforeShortcutEventArgs e)
		{
			if (BeforeShortcut != null) BeforeShortcut(this, e);
		}

		/// <summary>
		/// Fires the BeforeNavigateEvent
		/// </summary>
		/// <param name="e">Cancelable event args</param>
		public void OnBeforeNavigate(onlyconnect.BeforeNavigateEventArgs e) 
		{
			if (BeforeNavigate != null) BeforeNavigate(this, e);
		}

		#endregion Invoke Events ===============

		#region Setup Event Management ======================================================


		void ReadyStateChangeActions()
		{
			
			//defensive - I've known this to be called
			//after doc was deactivated
			if (this.m_htmldoc == null) return;

			mReadyState = this.m_htmldoc.readyState;
	
			if (mReadyState == "complete")
			{
				//if changed to "COMPLETE", set edit designer
				
				if (this.bLoadDocumentWhenReady)
				{
					this.LoadDocument(string.Empty);
				}
				else
				{
					if (this.IsDesignMode)
					{
						this.SetEditDesigner();
					}
				}

				if (this.bSetComposeSettingsWhenReady)
				{
					this.setDefaultFont();
				}

				//set HTMLEvents
				this.SetHTMLEvents();
				
			}
			
			//invoke ready state changed event
			this.InvokeReadyStateChanged(mReadyState);
			
		}



		public void SetEditDesigner()
		{
			this.m_htmldoc.designMode = "On";
			onlyconnect.IServiceProvider isp = (onlyconnect.IServiceProvider)m_htmldoc;
			onlyconnect.IHTMLEditServices es;
			System.Guid IHtmlEditServicesGuid = new System.Guid("3050f663-98b5-11cf-bb82-00aa00bdce0b");
			System.Guid SHtmlEditServicesGuid = new System.Guid(0x3050f7f9,0x98b5,0x11cf,0xbb,0x82,0x00,0xaa,0x00,0xbd,0xce,0x0b);
			IntPtr ppv;
			onlyconnect.IHTMLEditDesigner ds = (onlyconnect.IHTMLEditDesigner)site;
			if (isp != null)
			{
				isp.QueryService(ref SHtmlEditServicesGuid,ref IHtmlEditServicesGuid,out ppv);
				es = (onlyconnect.IHTMLEditServices)Marshal.GetObjectForIUnknown(ppv);
				int retval = es.AddDesigner(ds); 
				Marshal.Release(ppv);
			}
		}

		void SetDocEvents()
		{
			this.SetPropertyNotifyEvent();
	
		}

		void SetPropertyNotifyEvent()
		{
	
			if (m_htmldoc != null)
			{
				//get the connection point container
				UCOMIConnectionPointContainer icpc = (UCOMIConnectionPointContainer)m_htmldoc;

				//find the source interface
				//get IPropertyNotifySink interface
				Guid g = new Guid("9BFBBC02-EFF1-101A-84ED-00AA00341D07");

				icpc.FindConnectionPoint(ref g,out icp);

				//pass a pointer to the host to the connection point
				icp.Advise(this,out cookie);
						
			}
		}


		void SetHTMLEvents()
		{
			if ((m_htmldoc != null) && (HtmlEvent != null))
			{
				//m_htmldoc.onkeydown = this; 
				//keydown does not work, kepress does...
				m_htmldoc.onkeypress = this;
				m_htmldoc.onclick = this;
				m_htmldoc.onfocusin = this;
				m_htmldoc.onfocusout = this;
				m_htmldoc.onselectionchange = this;
			}
		}
		#endregion Setup Event Management ================

		#region private or internal methods

		private bool execCommand(int iCommand,object argument,bool bPromptUser, bool checkReadystate)
		{
			if (this.m_htmldoc == null)
			{
				return false;
			}
				
			//Check readystate

			if (checkReadystate)
			{
				if (this.ReadyState != "complete")
				{
					//throw new HtmlEditorException("Document is not ready");
					return false;
				}
			}

			//get the command target
			IOleCommandTarget ct = (IOleCommandTarget)this.m_htmldoc;

			if (ct == null)
			{
				throw new HtmlEditorException("Cannot get COM command target");
			}

			//exec the command
			System.Guid pguidCmdGroup = new Guid("DE4BA900-59CA-11CF-9592-444553540000");
			
			Object[] pvaOut = null;
			int iRetval;

			int iPromptUser;
			
			if (bPromptUser)
			{
				iPromptUser = (int)OLECMDEXECOPT.OLECMDEXECOPT_PROMPTUSER;
			}
			else
			{
				iPromptUser = (int)OLECMDEXECOPT.OLECMDEXECOPT_DONTPROMPTUSER;
			}
					
			object[] args = new object[] { argument };
			iRetval = ct.Exec(ref pguidCmdGroup,iCommand,iPromptUser,args,pvaOut);
		
			return (iRetval == 0);
	
		}

		/// <summary>
		/// Queries the status of the specified command
		/// </summary>
		private CommandStatus GetCommandInfo(int command) 
		{
			//get the command target
			IOleCommandTarget ct = (IOleCommandTarget)this.m_htmldoc;

			if (ct == null)
			{
				throw new HtmlEditorException("Cannot get COM command target");
			}
			
			System.Guid pguidCmdGroup = new Guid("DE4BA900-59CA-11CF-9592-444553540000");
			
			//Query the command target for the command status

			OLECMD oleCommand = new OLECMD();

			oleCommand.cmdID = command;
			
			int hr = ct.QueryStatus(ref pguidCmdGroup, 1, oleCommand, 0);
			
			if (hr != HRESULT.S_OK)
			{
				return CommandStatus.Unknown;
			}

		
			if ((oleCommand.cmdf | (int)OLECMDF.OLECMDF_LATCHED ) == (int) OLECMDF.OLECMDF_LATCHED)
			{
				return CommandStatus.EnabledAndToggledOn;
			}
			else if ((oleCommand.cmdf | (int)OLECMDF.OLECMDF_ENABLED) == (int)OLECMDF.OLECMDF_ENABLED)
			{
				return CommandStatus.Enabled;
			}
			else if ((oleCommand.cmdf | (int)OLECMDF.OLECMDF_SUPPORTED) == (int)OLECMDF.OLECMDF_SUPPORTED)
			{
				return CommandStatus.Disabled;
			}
			else 
			{
				return CommandStatus.Unsupported;
			}

		}

		private bool isCommandEnabled(int command)
		{
			CommandStatus cs = this.GetCommandInfo(command);

			if ((cs == CommandStatus.Enabled) | (cs == CommandStatus.EnabledAndToggledOn))
			{
				return true;
			}
			else
			{
				return false;
			}

		
		}

		
		private Boolean isCreated
		{
			get 
			{ 
				return (site != null) && (site.Document != null); 
			}
		}

		
		private void ActivateTimer_Tick(object sender, System.EventArgs e)
		{
			this.ActivateTimer.Enabled = false;
			site.ActivateDocument();
		}

		/// <summary>
		/// Executes The short cut keys that should be available and handles all of the cases of design mode versus not.
		/// </summary>
		/// <param name="Key">The key to process.</param>
		/// 
		private void doShortCut(Keys Key) 
		{
			
			//fire the BeforeShortcut event and cancel if necessary
			BeforeShortcutEventArgs e = new BeforeShortcutEventArgs(Key);
			this.OnBeforeShortcut(e);

			if (e.Cancel) return;
			
			switch(Key) 
			{
				case Keys.A: //Select All
					this.SelectAll();
					break;
				case Keys.C: //Copy
							Debug.WriteLine("Doing copy for shortcut");
					this.Copy();
					break;
				case Keys.P: //Print
					this.Print(false);
					break;
				case Keys.F: //Find
					this.ShowFindDialog();
					break;
			}

			if (this.IsDesignMode) 
			{
				//Only do the keys that work in design mode.
				switch(Key) 
				{
					case Keys.OemOpenBrackets: //Outdent
						this.Document.execCommand("Outdent", false, null);
						break;
					case Keys.OemCloseBrackets: //Indent
						this.Document.execCommand("Indent", false, null);
						break;
					case Keys.B: //Bold
						this.SetSelectionBold();
						break;
					case Keys.I: //Italic
						this.SetSelectionItalic();
						break;
					case Keys.U: //Underline
						this.Document.execCommand("underline", false, null);
						break;
					case Keys.L: //Left Justify
						this.Document.execCommand("JustifyLeft", false, null);
						break;
					case Keys.E: //Center Justify
						this.Document.execCommand("JustifyCenter", false, null);
						break;
					case Keys.R: //Right Justify
						this.Document.execCommand("JustifyRight", false, null);
						break;
					case Keys.V: //Paste
						this.Paste();
						break;
					case Keys.X: //Cut
						this.Cut();
						break;
					case Keys.Y: //Redo
						this.Redo();
						break;
					case Keys.Z: //Undo
						this.Undo();
						break;
				}
			}

		}

		private static uint GetCsColor(Color col)
		{
			return (uint) (col.R + (col.G * 256) + (col.B * 65536));
		}

		internal bool setDefaultFont()
		{
		
			if (!this.mDesignMode) return false;

			if (this.m_htmldoc.readyState != "complete")
			{
				this.bSetComposeSettingsWhenReady = true;
				return false;
			}

            this.bSetComposeSettingsWhenReady = false;
            
			object o = mComposeSettings.CommandString;

			if (this.execCommand(ComSupport.IDM_HTMLEDITMODE,true,false,false))
			{
				return this.execCommand(ComSupport.IDM_COMPOSESETTINGS,o,false,false);
			}
			else
			{
				return false;
			}
		}

		#endregion private or internal methods

		#region The main public API - eg. Copy, Paste, Print etc

		#region Load and save documents

		public void ReloadMshtml()
		{
			CleanupControl();
			initMshtml();
		}



		public void LoadDocument(String documentVal)
		{
			if ((documentVal != string.Empty) | (! this.bLoadDocumentWhenReady))
			{
				//if doc is waiting to load, it is already in string variable
				sDocument = documentVal;
				this.bLoadDocumentWhenReady = false;
			}
			else
			{
				this.bLoadDocumentWhenReady = false;
				this.iLoadAttempts += 1;
			}

			if (!isCreated)
			{
				throw new HtmlEditorException("Document not created");
			}

			if ( (this.m_htmldoc.readyState.ToLower() != "complete") & (this.m_htmldoc.readyState.ToLower() != "interactive"))
				//try to load on interactive as well as complete
			
			{
				if (iLoadAttempts < 2)
				{
					this.bLoadDocumentWhenReady = true;
					return;
				}
				else
				{
					throw new HtmlEditorException("Document not ready");
				}
			}

			//this is a fix for exception raised in UpdateUI
			site.mFullyActive = false;

			utils.LoadDocument(ref this.m_htmldoc,sDocument,this.mAlwaysLoadAnsi, false);
						
			this.iLoadAttempts = 0;
	
		}

		public void LoadUrl(String url)
		{
			this.url = url;
			
			if (!isCreated) return;

			//this is a workaround for a problem calling Caret.SetLocation before it
			//is ready, in UpdateUI
			site.mFullyActive = false;

			//don't ask for downloadOnly since clientsite is already set
			utils.LoadUrl(ref m_htmldoc,url,false);

		}

		public String GetDocumentSource()
		{
			if (!isCreated) return null;

			HTMLDocument  thedoc = this.Document;

			if (thedoc == null) return null;

			return utils.GetDocumentSource(ref thedoc,this.mDocumentEncoding);
			
		}

		#endregion Load and save documents

		[Browsable(false)]
		public mshtml.IHTMLElement CurrentElement
		{
			get
			{
				return mcurrentElement;
			}
		}

		[System.ComponentModel.TypeConverter(typeof(System.ComponentModel.ExpandableObjectConverter)),
	    DesignerSerializationVisibility(DesignerSerializationVisibility.Content),
		Description("Sets the initial font used by the editor. Disabled by default as it inserts a FONT tag.")]
		public ComposeSettings DefaultComposeSettings
		{
			get
			{
				if (this.mComposeSettings == null)
				{
					mComposeSettings = new ComposeSettings(this);
				}

				return mComposeSettings;
			}
		}

		[Description("Encoding to use when getting the string value of the HTML document"), DefaultValue(false)]
		public EncodingType DocumentEncoding
		{
			get {return mDocumentEncoding;}
			set {mDocumentEncoding = value;}
		}

		[Description("Turn off auto-detection of string type in LoadDocument and always treat it as an Ansi string."), DefaultValue(false)]
		public Boolean IsAnsiStreamAlwaysUsed
		{
			get
			{
				return this.mAlwaysLoadAnsi;
			}
			set
			{
				this.mAlwaysLoadAnsi = value;
			}
		}

		[Description("Is the control in edit mode so that users can edit the contents."), DefaultValue(false)]
		public Boolean IsDesignMode
		{
			get {return mDesignMode;}
			set 
			{
				if (value)
				{
					mDesignMode = true;
					if (this.m_htmldoc != null) 
					{
						this.m_htmldoc.designMode = "On";
						if (this.mComposeSettings.Enabled)
						{
							this.setDefaultFont();
						}
						this.LoadDocument("<html></html>");
					}
				}
				else
				{
					mDesignMode = false;
					if (this.m_htmldoc != null) 
					{
						
						this.m_htmldoc.designMode = "Off";
						
					}

				}
			}
		}

		[Description("Gets/Sets if the context menu on right clicks will be displayed."), DefaultValue(true)]
		public bool IsContextMenuEnabled
		{
			get 
			{
				return mIsContextMenuEnabled;
			}
			set
			{
				mIsContextMenuEnabled = value;
			}
		}

		[Description("Gets/Sets if the control can receive the focus."), DefaultValue(true)]
		public bool IsActivationEnabled
		{
			get 
			{
				return mAllowActivation;
			}
			set
			{
				mAllowActivation = value;
			}
		}

		[Description("Gets/Sets where the control stores preferences in the registry."), DefaultValue(null)]
		public string OptionKeyPath
		{
			get 
			{
				return this.mOptionKeyPath;
			}
			set
			{
				mOptionKeyPath = value;
			}
		}

		[Description("Gets/Sets if scripts, Java etc is enabled"), DefaultValue(true)]
		public bool IsActiveContentEnabled
		{
			//This must be set BEFORE the document is created

			get
			{
				return mEnableActiveContent ;
			}
		
			set
			{
				mEnableActiveContent = value;
			}
		}

		[Browsable(false)]
		public HTMLDocument Document
		{
			get
			{
				if (site == null)
				{
					return null;
				}

				if (site.Document != null) 
				{
					return (HTMLDocument)site.Document;
				}
				else
				{
					return null;
				}
			}
		}

		/// <summary>
		/// Gets/Sets if links that are clicked on in the editor will be opened in the editor or launch your default browser.
		/// </summary>
		[Description("Gets/Sets if links that are clicked on in the editor will be opened in the editor or launch your default browser."), DefaultValue(false)]
		public bool OpenLinksInNewWindow 
		{
			get {return mLinksNewWindow;}
			set 
			{
				mLinksNewWindow = value;
			}
		}

		public void setStyleSheet(string sFileName)
		{
			if (this.m_htmldoc == null)
			{
				return;
			}
	
			if (this.m_htmldoc.readyState.ToLower() == "complete")
			{
				m_htmldoc.createStyleSheet(sFileName,0);
			}

		}

		[Description("Gets ReadyState state of the control, eg loading, complete, etc"), Browsable(false)]
		public string ReadyState
		{
			get
			{
				if (this.m_htmldoc == null)
				{
					return String.Empty;
				}

				IHTMLDocument2 htmldoc = (IHTMLDocument2)this.m_htmldoc;

				return htmldoc.readyState.ToLower();
			}
		}
		
		public bool Copy()
		{
			return this.execCommand(ComSupport.IDM_COPY,null,false,true);
		}

		public bool Paste()
		{
			return this.execCommand(ComSupport.IDM_PASTE,null,false,true);
		}

		public bool SaveAs(string DefaultPath)
		{
			return this.execCommand(ComSupport.IDM_SAVEAS,DefaultPath,true,true);
		}

		public bool Undo()
		{
			return this.execCommand(ComSupport.IDM_UNDO,null,false,true);
		}

		public bool Redo()
		{
			return this.execCommand(ComSupport.IDM_REDO,null,false,true);
		}

		public bool Cut()
		{
			return this.execCommand(ComSupport.IDM_CUT,null,false,true);
		}

		public bool SetSelectionBold()
		{
			return this.execCommand(ComSupport.IDM_BOLD,null,false,true);
		}

		public bool SetSelectionItalic()
		{
			return this.execCommand(ComSupport.IDM_ITALIC,null,false,true);
		}

		public bool ClearSelectionFormatting()
		{
			return this.execCommand(ComSupport.IDM_REMOVEFORMAT,null,false,true);
		}

		[Browsable(false)]
		public Color SelectionBackColor
		{
			get
			{
				if (Document == null) return Color.Empty;
				else
				{
					object Result = Document.queryCommandValue("ForeColor");
					try 
					{
						return System.Drawing.ColorTranslator.FromHtml(Result.ToString());
					} 
					catch 
					{
						return Color.Empty;
					}
				}
			}
			set
			{
				//need to send a COLORREF value
				this.execCommand(ComSupport.IDM_BACKCOLOR,HtmlEditor.GetCsColor(value),false,true);
			}
		}

		/// <summary>
		/// Gets/Sets the color for the selected text.
		/// </summary>
		[Description("Gets/Sets the color of the selected text."), Browsable(false)]
		public Color SelectionForeColor 
		{
			get 
			{
				if (Document == null) return Color.Empty;
				else
				{
					object Result = Document.queryCommandValue("ForeColor");
					try 
					{
						return System.Drawing.ColorTranslator.FromHtml(Result.ToString());
					} 
					catch 
					{
						return Color.Empty;
					}
				}
			}
			set 
			{
				//need to send a COLORREF value
				this.execCommand(ComSupport.IDM_FORECOLOR,HtmlEditor.GetCsColor(value),false,true);
			}
		}

		[Description("Gets/sets the font for the selected text"), Browsable(false)]
		public Font SelectionFont
		{
			get 
			{
				try 
				{ 
					//This is here because of hosting issues on a UserControl 
					if (this.Document == null) return null;

					System.Drawing.FontStyle fs = new FontStyle();
					int FontSize = 8;

					if (Convert.ToBoolean(Document.queryCommandValue("bold"))) fs |= FontStyle.Bold;
					if (Convert.ToBoolean(Document.queryCommandValue("italic"))) fs |= FontStyle.Italic;
					if (Convert.ToBoolean(Document.queryCommandValue("underline"))) fs |= FontStyle.Underline;

					switch(Convert.ToInt32(Document.queryCommandValue("FontSize"))) 
					{
						case 1:
							FontSize = 8;
							break;
						case 2:
							FontSize = 10;
							break;
						case 3:
							FontSize = 12;
							break;
						case 4:
							FontSize = 18;
							break;
						case 5:
							FontSize = 24;
							break;
						case 6:
							FontSize = 36;
							break;
						case 7:
							FontSize = 48;
							break;
					}

					string FontName = Document.queryCommandValue("FontName").ToString();

					return new Font(FontName, FontSize, fs);
				} 
				catch 
				{
					return null;
				}
			}
			set 
			{
				try 
				{ //This is here because of hosting issues on a UserControl
					if ((Document == null) | (value == null)) return;

					Document.execCommand("FontName", false, value.Name);

					if (value.SizeInPoints <= 8) 
					{
						Document.execCommand("FontSize", false, 1);
					} 
					else if (value.SizeInPoints <= 10) 
					{
						Document.execCommand("FontSize", false, 2);
					} 
					else if (value.SizeInPoints <= 12) 
					{
						Document.execCommand("FontSize", false, 3);
					} 
					else if (value.SizeInPoints <= 18) 
					{
						Document.execCommand("FontSize", false, 4);
					} 
					else if (value.SizeInPoints <= 24) 
					{
						Document.execCommand("FontSize", false, 5);
					} 
					else if (value.SizeInPoints <= 36) 
					{
						Document.execCommand("FontSize", false, 6);
					} 
					else 
					{
						Document.execCommand("FontSize", false, 7);
					}

					if (Convert.ToBoolean(Document.queryCommandValue("bold")) != Font.Bold) Document.execCommand("bold", false, null);
					if (Convert.ToBoolean(Document.queryCommandValue("italic")) != Font.Bold) Document.execCommand("italic", false, null);
					if (Convert.ToBoolean(Document.queryCommandValue("underline")) != Font.Bold) Document.execCommand("underline", false, null);
				} 
				catch {}
			}
		}

		/// <summary>
		/// Gets/Sets the alignment of the selected text.
		/// </summary>
		[Description("Gets/Sets the alignment of the selected text."), Browsable(false)]
		public System.Windows.Forms.HorizontalAlignment SelectionAlignment 
		{
			get 
			{
				if (Document == null) return HorizontalAlignment.Left;
				else
				{
					if (Convert.ToBoolean(Document.queryCommandValue("JustifyRight"))) 
					{
						return HorizontalAlignment.Right;
					} 
					else if (Convert.ToBoolean(Document.queryCommandValue("JustifyCenter"))) 
					{
						return HorizontalAlignment.Center;
					} 
					else 
					{
						return HorizontalAlignment.Left;
					}
				}
			}
			set 
			{
				if (Document == null) return;

				switch (value) 
				{
					case HorizontalAlignment.Left:
						Document.execCommand("JustifyLeft", false, null);
						break;
					case HorizontalAlignment.Center:
						Document.execCommand("JustifyCenter", false, null);
						break;
					case HorizontalAlignment.Right:
						Document.execCommand("JustifyRight", false, null);
						break;
				}
			}
		}

		/// <summary>
		/// Get/Sets if numbering is on for the selected text.
		/// </summary>
		[Description("Get/Sets if numbering is on for the selected text."), Browsable(false)]
		public bool SelectionNumbering 
		{
			get 
			{
				if (Document == null) return false;
				else return Convert.ToBoolean(Document.queryCommandValue("InsertOrderedList"));
			} 
			set 
			{
				if (Document != null && Convert.ToBoolean(Document.queryCommandValue("InsertOrderedList")) != value) Document.execCommand("InsertOrderedList", false, null);
			}
		}

		/// <summary>
		/// Gets/Sets if bullets are on or off for the selected text.
		/// </summary>
		[Description("Gets/Sets if bullets are on or off for the selected text."), Browsable(false)]
		public bool SelectionBullets 
		{
			get 
			{
				if (Document == null) return false;
				else return Convert.ToBoolean(Document.queryCommandValue("InsertUnorderedList"));
			}
			set 
			{
				if (Document != null && Convert.ToBoolean(Document.queryCommandValue("InsertUnorderedList")) != value) Document.execCommand("InsertUnorderedList", false, null);
			}
		}

		[Description("Select the entire document"), Browsable(false)]
			public bool SelectAll()
		{
			return this.execCommand(ComSupport.IDM_SELECTALL,null,false,true);
		}

		public void ClearSelection()
		{
			this.Document.selection.empty();
		}
        
		public void DeleteSelection()
		{
			this.execCommand(ComSupport.IDM_DELETE,null,false,true);
		}

		public void ShowFindDialog()
		{
			this.execCommand(ComSupport.IDM_FIND,null,true,true);
		}

		public void Print(bool bPreview)
		{
			this.Print(bPreview,string.Empty,true);

		}

		public void Print(bool bPreview, String sTemplatePath)
		{
			this.Print(bPreview,string.Empty,true);

		}

		public void Print(bool bPreview,String sTemplatePath,bool bPromptUser)
		{
			Object pvaIn;

			if (sTemplatePath == string.Empty)
			{
				pvaIn = null;
			}
			else if (! File.Exists(sTemplatePath))
			{
				pvaIn = null;
			}
			else
			{
				pvaIn = sTemplatePath;
			}

							
			if (bPreview)
			{
				this.execCommand (ComSupport.IDM_PRINTPREVIEW,pvaIn,bPromptUser,true);
		
			}
			else
			{
				this.execCommand (ComSupport.IDM_PRINT,pvaIn,bPromptUser,true);
			}


		}

		public bool Stop()
		{
			return this.execCommand(ComSupport.IDM_STOP,null,false,false);
		}

		#endregion Main public API ==========================

		#region IPropertyNotifySink -- notifications of property changes =====================

		//implementation of IPropertyNotifySink -- notifications of property changes
		/*
			 * The IPropertyNotifySink interface is implemented 
			 * by a sink object 
			 * to receive notifications about property changes 
			 * from an object that supports IPropertyNotifySink as an "outgoing" interface. 
			 * */

		public int OnChanged(int iDispID)
		{
	
			if (iDispID == ComSupport.DISPID_READYSTATE)
			{
				this.ReadyStateChangeActions ();
			}

			if (iDispID == ComSupport.DISPID_UNKNOWN)
				//true if more than one property has changed
			{
		
				if (this.m_htmldoc.readyState != this.mReadyState)
				{
					this.ReadyStateChangeActions ();
				}

			
			}
			return HRESULT.S_OK; 
		}

		public int OnRequestEdit(int iDispID)
		{
			return HRESULT.S_OK; //indicates change is allowed
		}

		#endregion IPropertyNotifySink ===============

	}

	class HandleWndProc: NativeWindow
	{
		internal HtmlEditor thecontrol;

		protected override void WndProc(ref Message message)
		{
			//The idea is only pass the key presses and mouse clicks (and not the right clicks) to the base form to process the events correctly.
			if ((message.Msg >= 0x100 && message.Msg <= 0x0108) || (message.Msg >= 0x200 && message.Msg <= 0x020A && message.Msg != ComSupport.WM_LBUTTONDOWN && message.Msg != ComSupport.WM_RBUTTONDOWN && message.Msg != ComSupport.WM_LBUTTONDBLCLK && message.Msg != ComSupport.WM_MBUTTONDBLCLK && message.Msg != ComSupport.WM_RBUTTONDBLCLK)) 
			{
				thecontrol.InvokeWndProc(ref message);
			} 
			else 
			{
				//These are separate because if you pass it to the base control then the right mouse clicks etc. don't fire.
				switch(message.Msg) 
				{
					case ComSupport.WM_LBUTTONDOWN:
						thecontrol.InvokeOnMouseDown(new MouseEventArgs(MouseButtons.Left, 1, message.LParam.ToInt32() & 0xffff, Convert.ToInt32((message.LParam.ToInt32() & 0xffff0000) >> 16), 0));
						break;
					case ComSupport.WM_RBUTTONDOWN:
						thecontrol.InvokeOnMouseDown(new MouseEventArgs(MouseButtons.Right, 1, message.LParam.ToInt32() & 0xffff, Convert.ToInt32((message.LParam.ToInt32() & 0xffff0000) >> 16), 0));
						break;
					case ComSupport.WM_LBUTTONDBLCLK:
					case ComSupport.WM_MBUTTONDBLCLK:
					case ComSupport.WM_RBUTTONDBLCLK:
						thecontrol.InvokeOnDoubleClick();
						break;
				}
			}

			base.WndProc(ref message);
		}
	}

	public enum EncodingType : byte
	{
		UTF7,
		UTF8,
		Unicode,
		ASCII,
		WindowsCurrent
	}

	public enum CommandStatus : byte
	{
		Unsupported,
		Disabled,
		Enabled,
		EnabledAndToggledOn,
		Unknown
	}
}
